#!/usr/bin/env python

import io

import imageio
import matplotlib as mpl
import numpy as np
import PIL
import scipy as sp
import scipy.fftpack
import seaborn as sns
from matplotlib import colors
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import Divider, Size, make_axes_locatable
from skimage.util import view_as_blocks, view_as_windows


def add_cbar(ax, plot, **kwargs):
    """Adds a colourbar whose height matches that of the axis."""
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad='2%')
    cb = plt.colorbar(plot, cax=cax, **kwargs)
    cax.tick_params(labelsize=25)
    return cb


def create_fig(figsize):
    """Creates a figure, axis and colorbar axis such that the main axis has the specified size after saving."""
    px = 1 / plt.rcParams['figure.dpi']
    sY, sX = figsize

    pad = sX * 0.05
    cbsz = sX * 0.05
    fig = plt.figure(figsize=((sX + pad + cbsz) * px, sY * px))

    # The first items are for padding and the second items are for the axes.
    # sizes are in inch.
    h = [Size.Fixed(0 * px), Size.Fixed(sX * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (0, 0, 1 - (pad + cbsz) / (sY + pad + cbsz), 1),
                      h,
                      v,
                      aspect=False)
    # The width and height of the rectangle are ignored.
    ax = fig.add_axes(divider.get_position(),
                      axes_locator=divider.new_locator(nx=1, ny=1))
    h = [Size.Fixed(0 * px), Size.Fixed(cbsz * px)]
    v = [Size.Fixed(0 * px), Size.Fixed(sY * px)]
    divider = Divider(fig, (1 - (pad + cbsz) / (sX + pad + cbsz), 0, 1, 1),
                      h,
                      v,
                      aspect=False)
    cax = fig.add_axes(divider.get_position(),
                       axes_locator=divider.new_locator(nx=1, ny=1))
    cax.tick_params(labelsize=25)
    return fig, ax, cax


def jpeg_compress(img, q=None, quality=None):
    assert q is not None or quality is not None
    pimg = PIL.Image.fromarray((img * 255).astype(np.uint8))
    out = io.BytesIO()
    if q is not None:
        pimg.save(out, format='JPEG', qtables={0: [q] * 64})
    else:
        pimg.save(out, format='JPEG', quality=quality)
    out.seek(0)
    jpeg = PIL.Image.open(out)
    quantization = jpeg.quantization
    #quantization = np.array(jpeg.quantization[0])
    jpeg = np.array(jpeg) / 255
    return jpeg, quantization


def get_noise_curve(img, q, ls):
    in_σs = ls * q
    out_σ = []
    Y, X = img.shape
    for i, in_σ in enumerate(in_σs):
        noise = np.random.normal(0, in_σ, (Y, X)).round() / 255
        img_noisy = np.clip(img + noise, 0, 1)
        jpeg_noiseless, _ = jpeg_compress(img, q)
        jpeg_noisy, _ = jpeg_compress(img_noisy, q)
        residual = jpeg_noisy - jpeg_noiseless
        σ = (residual).std() * 255
        out_σ.append(σ)
    out_σ = np.array(out_σ)
    return in_σs, np.array(out_σ)


def series(x, l):
    return np.exp(-2 * x**2 * np.pi**2 * l**2) * (1 / (np.pi**2 * l**2) +
                                                  4 * x**2) * (-1)**l


def f(x):
    return x**2 + 1 / 12 + np.sum([series(x, l) for l in range(1, 1000)])


def plot_noise_curves(img):
    qs = [1, 8, 16, 32, 64, 128, 255]
    fig, ax = plt.subplots(figsize=(7, 6))
    lw = 2
    ls = np.linspace(.5, 1.3, 50)
    th = np.sqrt(np.maximum(0, [f(k) for k in ls]))
    ax.plot(ls, th, lw=lw, label="Model")
    for q in qs:
        in_σ, out_σ = get_noise_curve(img, q, ls)
        ax.plot(in_σ / q, out_σ / q, label=f"q={q}", lw=lw)
    ax.plot(in_σ / q, in_σ / q, '--', lw=1)
    plt.legend(fontsize=14)
    plt.xlabel("Input σ/q", fontsize=14)
    plt.ylabel("Post-compression σ/q", fontsize=14)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.savefig("noise_curve.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    plt.close()


def get_dct_std(img, qf, σ):
    Y, X = img.shape
    noise = np.random.normal(0, σ, (Y, X)).round() / 255
    img_noisy = np.clip(img + noise, 0., 1.)
    jpeg_noiseless, _ = jpeg_compress(img, quality=qf)
    jpeg_noisy, _ = jpeg_compress(img_noisy, quality=qf)
    residual = jpeg_noisy - img

    Y, X = residual.shape
    Y -= Y % 8
    X -= X % 8
    residual = residual[:Y, :X]
    blocks = view_as_blocks(residual, (8, 8))
    dct = (sp.fft.dctn(blocks, axes=(-1, -2), norm='ortho') /
           8).std(axis=(0, 1)) * 255
    fig, ax = plt.subplots(figsize=(6, 6))
    ax.axis('off')
    p = ax.matshow(dct, cmap='mako', vmin=0, vmax=1.5)
    add_cbar(ax, p, extend='max')
    plt.savefig(f'dct_qf{qf}_std{σ}.png',
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    plt.close()
    fig, ax, cax = create_fig(img.shape)
    ax.axis('off')
    p = ax.matshow(np.abs(residual * 255), vmin=0, vmax=25, cmap='mako')
    plt.colorbar(p, cax=cax, extend='max')
    plt.savefig(f'residual_qf{qf}_std{σ}.png',
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    plt.close()
    # imageio.imwrite(f'residual_qf{qf}_σ{σ}.png',
    # np.abs(residual * 255).astype(np.uint8))


if __name__ == "__main__":
    img = imageio.imread("input_0.png") / 255
    if img.ndim == 3:
        img = .299 * img[:, :, 0] + .587 * img[:, :, 1] + .114 * img[:, :, 2]
    plot_noise_curves(img)
    for qf in [10, 50, 70, 90, 100]:
        for σ in [1, 3, 10]:
            get_dct_std(img, qf, σ)
